/*_############################################################################
  _##
  _##  SNMP4J-AgentX - BufferedMOTableModel.java
  _##
  _##  Copyright (C) 2005-2014  Frank Fock (SNMP4J.org)
  _##
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  _##  MA  02110-1301  USA
  _##
  _##########################################################################*/

package org.snmp4j.agent.mo;

import org.snmp4j.mp.SnmpConstants;
import org.snmp4j.smi.OID;
import org.snmp4j.smi.Variable;

import java.util.*;

/**
 * The BufferedMOTableModel is a read-only table model that dynamically loads data from an
 * backend as needed.
 *
 * @author Frank Fock
 */
public abstract class BufferedMOTableModel<R extends MOTableRow> implements MOTableModel<R> {

  protected SortedMap<OID, BufferedMOTableRow<R>> bufferedRows =
      Collections.synchronizedSortedMap(new TreeMap<OID, BufferedMOTableRow<R>>());
  protected BufferedMOTableRow<R> firstRow;
  protected BufferedMOTableRow<R> lastRow;

  protected MOTableRowFactory<R> rowFactory;

  protected int columnCount = 0;
  protected long refreshIntervalNanoSeconds = 5L * SnmpConstants.MILLISECOND_TO_NANOSECOND * 1000L;

  protected int chunkSize = 10;

  protected BufferedMOTableModel(MOTableRowFactory<R> rowFactory) {
    this.rowFactory = rowFactory;
  }

  /**
   * Sets the factory instance to be used for creating rows for this model.
   *
   * @param rowFactory
   *    a <code>MOTableRowFactory</code> instance or <code>null</code> to
   *    disable row creation.
   */
  public void setRowFactory(MOTableRowFactory<R> rowFactory) {
    this.rowFactory = rowFactory;
  }

  @Override
  public int getColumnCount() {
    return columnCount;
  }

  @Override
  public abstract int getRowCount();

  @Override
  public abstract boolean containsRow(OID index);

  @Override
  public R getRow(OID index) {
    BufferedMOTableRow<R> row = getRowFromBuffer(index);
    if (row == null) {
      Variable[] rowValues = fetchRow(index);
      if (rowValues != null) {
        R r = rowFactory.createRow(index, rowValues);
        row = new BufferedMOTableRow<R>(r);
      }
      else {
        row = new BufferedMOTableRow<R>(null);
      }
      bufferedRows.put(index, row);
    }
    return row.getBufferedRow();
  }

  @Override
  public Iterator<R> iterator() {
    return tailIterator(null);
  }

  @Override
  public Iterator<R> tailIterator(OID lowerBound) {
    return new RowBufferIterator(lowerBound);
  }

  @Override
  public abstract OID lastIndex();

  @Override
  public abstract OID firstIndex();

  @Override
  public R firstRow() {
    if ((firstRow != null) && (isRowValid(firstRow.getLastRefresh()))) {
      return firstRow.getBufferedRow();
    }
    OID firstIndex = firstIndex();
    if (firstIndex != null) {
      R row = getRow(firstIndex);
      if (row != null) {
        this.firstRow = getRowFromBuffer(firstIndex);
      }
      return row;
    }
    return null;
  }

  @Override
  public R lastRow() {
    if ((lastRow != null) && (isRowValid(lastRow.getLastRefresh()))) {
      return lastRow.getBufferedRow();
    }
    OID lastIndex = lastIndex();
    if (lastIndex != null) {
      R row = getRow(lastIndex);
      if (row != null) {
        this.firstRow = getRowFromBuffer(lastIndex);
      }
      return row;
    }
    return null;
  }

  protected boolean isRowValid(long lastRefreshNanoTime) {
    return (System.nanoTime() - lastRefreshNanoTime < refreshIntervalNanoSeconds);
  }

  protected BufferedMOTableRow<R> getRowFromBuffer(OID index) {
    BufferedMOTableRow<R> row = bufferedRows.get(index);
    if (row != null) {
      if (!isRowValid(row.getLastRefresh())) {
        bufferedRows.remove(index);
        return null;
      }
      return row;
    }
    return null;
  }

  /**
   * Fetches the specified row from the backend source.
   * @param index
   *    the rows index OID value.
   * @return
   *    the values of the fetched row or <code>null</code> if the row does not exists.
   */
  protected abstract Variable[] fetchRow(OID index);

  /**
   * Fetches a list of rows from the backend source.
   * @param lowerBound
   *   the lower bound index (inclusive) of the first row to return.
   * @param chunkSize
   *   the maximum number of rows to return. Less rows may be returned even if there are more
   *   available.
   * @return
   *   a list of rows fetched from the backend. If the last element of the returned list is
   *   <code>null</code>, no more rows are available at the source.
   */
  protected abstract List<R> fetchNextRows(OID lowerBound, int chunkSize);

  protected void removeRow(OID index) {
    throw new UnsupportedOperationException();
  }

  protected class RowBufferIterator implements Iterator<R> {

    private OID currentIndex;
    private R nextRow;

    public RowBufferIterator(OID lowerBound) {
      this.currentIndex = lowerBound;
      if (currentIndex == null) {
        nextRow = firstRow();
      }
      else {
        nextRow = fetchNext(currentIndex);
      }
    }

    @Override
    public boolean hasNext() {
      return nextRow != null;
    }

    @Override
    public R next() {
      R next = nextRow;
      if (next != null) {
        nextRow = fetchNext(next.getIndex());
      }
      return next;
    }

    @Override
    public void remove() {
      removeRow(currentIndex);
    }

    protected R fetchNext(OID lowerBound) {
      OID lowerBoundInc = lowerBound.successor();
      SortedMap<OID,BufferedMOTableRow<R>> nextRows = bufferedRows.tailMap(lowerBoundInc);
      if (!nextRows.isEmpty()) {
        BufferedMOTableRow<R> bufferedRow = nextRows.values().iterator().next();
        if (isRowValid(bufferedRow.getLastRefresh())) {
          return bufferedRow.getBufferedRow();
        }
      }
      List<R> newBufferRows = fetchNextRows(lowerBoundInc, chunkSize);
      for (R row : newBufferRows) {
        bufferedRows.put(row.getIndex(), new BufferedMOTableRow<R>(row));
      }
      if (newBufferRows.isEmpty()) {
        return null;
      }
      return newBufferRows.iterator().next();
    }
  }

  protected class BufferedMOTableRow<R extends MOTableRow> implements MOTableRow<Variable> {

    private long lastRefresh;
    private R bufferedRow;

    public BufferedMOTableRow(R bufferedRow) {
      this.bufferedRow = bufferedRow;
      lastRefresh = System.nanoTime();
    }

    public R getBufferedRow() {
      return bufferedRow;
    }

    @Override
    public OID getIndex() {
      return bufferedRow.getIndex();
    }

    @Override
    public Variable getValue(int column) {
      return bufferedRow.getValue(column);
    }

    @Override
    public MOTableRow getBaseRow() {
      return bufferedRow.getBaseRow();
    }

    @Override
    public void setBaseRow(MOTableRow baseRow) {
      bufferedRow.setBaseRow(baseRow);
    }

    @Override
    public int size() {
      return bufferedRow.size();
    }

    public long getLastRefresh() {
      return lastRefresh;
    }

    public void setLastRefresh(long lastRefresh) {
      this.lastRefresh = lastRefresh;
    }
  }
}
